#include "FakePrimarySurface.h"

#include "BitmapSaver.h"

FakePrimarySurface::FakePrimarySurface(IDirectDraw4* ddraw) {
	Refs=1;
	MenuMode=true;
	HRESULT hr;

	memset(&rDesc, 0, sizeof(DDSURFACEDESC2));
	rDesc.dwSize=sizeof(DDSURFACEDESC2);
	rDesc.dwFlags=DDSD_CAPS;
#ifdef NONEXCLUSIVE
	rDesc.ddsCaps.dwCaps=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE;
#else
	rDesc.dwFlags|=DDSD_BACKBUFFERCOUNT;
	rDesc.ddsCaps.dwCaps=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX|DDSCAPS_3DDEVICE;
	rDesc.dwBackBufferCount=1;
#endif

	hr=ddraw->CreateSurface(&rDesc, &Real, 0);
	CHECKFAIL(hr);
	hr=Real->GetSurfaceDesc(&rDesc);
	CHECKFAIL(hr);

#ifdef NONEXCLUSIVE
	DDSURFACEDESC2 desc=rDesc;
	desc.dwFlags=DDSD_CAPS|DDSD_WIDTH|DDSD_HEIGHT|DDSD_PIXELFORMAT;
	desc.ddsCaps.dwCaps=DDSCAPS_VIDEOMEMORY|DDSCAPS_3DDEVICE;
	desc.dwWidth=gWidth;
	desc.dwHeight=gHeight;
	desc.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	desc.ddpfPixelFormat.dwFlags=DDPF_RGB;
	desc.ddpfPixelFormat.dwRBitMask=0xff0000;
	desc.ddpfPixelFormat.dwGBitMask=0xff00;
	desc.ddpfPixelFormat.dwBBitMask=0xff;
	desc.ddpfPixelFormat.dwRGBBitCount=32;

	hr=ddraw->CreateSurface(&desc, &Real2, 0);
#else
	DDSCAPS2 caps;
	memset(&caps, 0, sizeof(DDSCAPS2));
	caps.dwCaps=DDSCAPS_BACKBUFFER;
	hr=Real->GetAttachedSurface(&caps, &Real2);
	CHECKFAIL(hr);

	rDesc.dwFlags|=DDSD_BACKBUFFERCOUNT;
	rDesc.ddsCaps.dwCaps=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX|DDSCAPS_3DDEVICE;
	rDesc.dwBackBufferCount=1;
#endif

	rDesc.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	rDesc.ddpfPixelFormat.dwFlags=DDPF_RGB;
	rDesc.ddpfPixelFormat.dwRBitMask=0xf800;
	rDesc.ddpfPixelFormat.dwGBitMask=0x7e0;
	rDesc.ddpfPixelFormat.dwBBitMask=0x1f;
	rDesc.ddpfPixelFormat.dwRGBBitCount=16;

	memset(&mDesc, 0, sizeof(DDSURFACEDESC2));
	mDesc.dwSize=sizeof(DDSURFACEDESC2);
	mDesc.dwFlags=DDSD_CAPS|DDSD_PIXELFORMAT|DDSD_HEIGHT|DDSD_WIDTH;
	if(UseSysMemOverlay==1) {
		mDesc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;
	} else {
		mDesc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_VIDEOMEMORY;
	}
	mDesc.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	mDesc.ddpfPixelFormat.dwFlags=DDPF_RGB;
	mDesc.ddpfPixelFormat.dwRBitMask=0xf800;
	mDesc.ddpfPixelFormat.dwGBitMask=0x7e0;
	mDesc.ddpfPixelFormat.dwBBitMask=0x1f;
	mDesc.ddpfPixelFormat.dwRGBBitCount=16;
	mDesc.dwWidth=640;
	mDesc.dwHeight=480;

	oDesc=mDesc;
	oDesc.dwWidth=gWidth;
	oDesc.dwHeight=gHeight;
	if(UseSysMemOverlay==2) oDesc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;

	hr=ddraw->CreateSurface(&mDesc, &menu, 0);
	CHECKFAIL(hr);
	hr=menu->GetSurfaceDesc(&mDesc);
	CHECKFAIL(hr);
	mDesc.ddsCaps.dwCaps|=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX;
	mDesc.dwBackBufferCount=1;
	mDesc.dwFlags|=DDSD_BACKBUFFERCOUNT;

	hr=ddraw->CreateSurface(&oDesc, &overlay, 0);
	CHECKFAIL(hr);
	hr=overlay->GetSurfaceDesc(&oDesc);
	CHECKFAIL(hr);
	if(UseSysMemOverlay!=2) {
		DDCOLORKEY key;
		key.dwColorSpaceHighValue=TextOverlayColourKey;
		key.dwColorSpaceLowValue=TextOverlayColourKey;
		hr=overlay->SetColorKey(DDCKEY_SRCBLT, &key);
		CHECKFAIL(hr);
	}
	oDesc.ddsCaps.dwCaps|=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX;
	oDesc.dwBackBufferCount=1;
	oDesc.dwFlags|=DDSD_BACKBUFFERCOUNT;

	ClearOverlay();
}
void FakePrimarySurface::ClearOverlay() {
	DDBLTFX fx;
	memset(&fx, 0, sizeof(DDBLTFX));
	fx.dwSize=sizeof(DDBLTFX);
	fx.dwFillColor=TextOverlayColourKey;
	HRESULT hr=overlay->Blt(0, 0, 0, DDBLT_WAIT|DDBLT_COLORFILL, &fx);
	CHECKFAIL(hr);
}
void FakePrimarySurface::SetMenuMode(bool b) { MenuMode=b; GetAsyncKeyState(VK_SNAPSHOT); }
IDirectDrawSurface4* FakePrimarySurface::QueryReal() { return Real2; }
// IUnknown methods
HRESULT _stdcall FakePrimarySurface::QueryInterface(REFIID a, LPVOID * b) {
	//69c11c3e = IID_IDirectDrawGammaControl
	//93281502 = IID_IDirect3DTexture2
	if(a==IID_IDirectDrawGammaControl) {
		return Real->QueryInterface(a,b);
	} else UNUSEDFUNCTION;
}
ULONG _stdcall FakePrimarySurface::AddRef()  { return ++Refs; }
ULONG _stdcall FakePrimarySurface::Release() {
	if(!--Refs) {
		menu->Release();
		Real->Release();
		//Real2->Release(); //Complex surfaces release all resources on exit
		delete this;
		return 0;
	} else return Refs;
}
// IDirectDrawSurface methods
HRESULT _stdcall FakePrimarySurface::AddAttachedSurface(LPDIRECTDRAWSURFACE4 a) {
	return Real2->AddAttachedSurface(a);
}
HRESULT _stdcall FakePrimarySurface::AddOverlayDirtyRect(LPRECT) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::Blt(LPRECT,LPDIRECTDRAWSURFACE4, LPRECT,DWORD, LPDDBLTFX) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::BltBatch(LPDDBLTBATCH, DWORD, DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::BltFast(DWORD,DWORD,LPDIRECTDRAWSURFACE4, LPRECT,DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::DeleteAttachedSurface(DWORD a,LPDIRECTDRAWSURFACE4 b) {
	return Real2->DeleteAttachedSurface(a,b);
}
HRESULT _stdcall FakePrimarySurface::EnumAttachedSurfaces(LPVOID,LPDDENUMSURFACESCALLBACK2) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::EnumOverlayZOrders(DWORD,LPVOID,LPDDENUMSURFACESCALLBACK2) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::Flip(LPDIRECTDRAWSURFACE4 a, DWORD b) {
#define BITS16TO32(a) ( (((a)&0xf800)<<8) + (((a)&0x7e0)<<5) + (((a)&0x1f)<<3) )
	if(MenuMode) UNUSEDFUNCTION;
	HRESULT hr;
	if(UseSysMemOverlay==2) {
		DDSURFACEDESC2 realdesc;
		DDSURFACEDESC2 overdesc;
		realdesc.dwSize=overdesc.dwSize=sizeof(DDSURFACEDESC2);
		hr=overlay->Lock(0, &overdesc, DDLOCK_READONLY|DDLOCK_WAIT, 0);
		hr=Real2->Lock(0, &realdesc, DDLOCK_WAIT, 0);
		realdesc.lPitch/=4;
		overdesc.lPitch/=2;
		for(DWORD y=0;y<gHeight;y++) {
			for(DWORD x=0;x<gWidth;x++) {
				if(((WORD*)overdesc.lpSurface)[y*overdesc.lPitch+x]!=TextOverlayColourKey)
					((DWORD*)realdesc.lpSurface)[y*realdesc.lPitch+x]=BITS16TO32(((WORD*)overdesc.lpSurface)[y*overdesc.lPitch+x]);
			}
		}
		hr=Real2->Unlock(0);
		hr=overlay->Unlock(0);
	} else {
		//hr=Real2->Blt(0, overlay, 0, DDBLT_KEYSRC|DDBLT_WAIT, 0);
		hr=Real2->BltFast(0, 0, overlay, 0, DDBLTFAST_SRCCOLORKEY|DDBLTFAST_WAIT);
		CHECKFAIL(hr);
	}
	if(GetAsyncKeyState(VK_SNAPSHOT)) {
		/*DWORD total, free;
		DDSCAPS2 caps;
		memset(&caps, 0, sizeof(caps));
		caps.dwCaps=DDSCAPS_TEXTURE|DDSCAPS_VIDEOMEMORY|DDSCAPS_LOCALVIDMEM;
		HRESULT hr=rDDraw->GetAvailableVidMem(&caps, &total, &free);
		DWORD used=(total-free)/(1024*1024);
		total/=(1024*1024);
		free/=(1024*1024);*/
		DDSURFACEDESC2 desc;
		memset(&desc, 0, sizeof(desc));
		desc.dwSize=sizeof(desc);
		Real2->Lock(0, &desc, DDLOCK_READONLY|DDLOCK_WAIT, 0);
		SaveBitmap(desc.lpSurface, desc.lPitch);
		Real2->Unlock(0);
	}
	ClearOverlay();
#ifdef NONEXCLUSIVE
	hr=Real->BltFast(0,0,Real2,0,DDBLTFAST_WAIT);
#else
	hr=Real->Flip(a,b);
	CHECKFAIL(hr);
#endif
	return DD_OK;
}
HRESULT _stdcall FakePrimarySurface::GetAttachedSurface(LPDDSCAPS2 a, LPDIRECTDRAWSURFACE4 * b) {
	if(a->dwCaps&DDSCAPS_BACKBUFFER) {
		AddRef();
		*b=this;
		return DD_OK;
	} else UNUSEDFUNCTION;
}
HRESULT _stdcall FakePrimarySurface::GetBltStatus(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetCaps(LPDDSCAPS2 a) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetClipper(LPDIRECTDRAWCLIPPER *) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetColorKey(DWORD, LPDDCOLORKEY) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetDC(HDC *) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetFlipStatus(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetOverlayPosition(LPLONG, LPLONG) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetPalette(LPDIRECTDRAWPALETTE *) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetPixelFormat(LPDDPIXELFORMAT) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetSurfaceDesc(LPDDSURFACEDESC2 a) {
	*a=MenuMode?mDesc:oDesc;
	return DD_OK;
}
HRESULT _stdcall FakePrimarySurface::Initialize(LPDIRECTDRAW, LPDDSURFACEDESC2) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::IsLost() {
	return Real->IsLost();
}
HRESULT _stdcall FakePrimarySurface::Lock(LPRECT a,LPDDSURFACEDESC2 b,DWORD c,HANDLE d) {
	if(MenuMode) {
		HRESULT hr=menu->Lock(a,b,c,d);
		CHECKFAIL(hr);
		if(FAILED(hr)) return hr;
		b->ddsCaps.dwCaps|=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX;
		b->dwBackBufferCount=1;
		b->dwFlags|=DDSD_BACKBUFFERCOUNT;
		return DD_OK;
	} else {
		HRESULT hr=overlay->Lock(a,b,c,d);
		CHECKFAIL(hr);
		if(FAILED(hr)) return hr;
		return DD_OK;
	}
}
HRESULT _stdcall FakePrimarySurface::ReleaseDC(HDC) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::Restore() {
	switch(UseSysMemOverlay) {
		case 0: overlay->Restore();
		case 2: menu->Restore();
	}
	return Real->Restore();
}
HRESULT _stdcall FakePrimarySurface::SetClipper(LPDIRECTDRAWCLIPPER) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::SetColorKey(DWORD, LPDDCOLORKEY) { UNUSEDFUNCTION;	}
HRESULT _stdcall FakePrimarySurface::SetOverlayPosition(LONG, LONG) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::SetPalette(LPDIRECTDRAWPALETTE) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::Unlock(LPRECT a) {
	if(MenuMode) {
		HRESULT hr=menu->Unlock(a);
		CHECKFAIL(hr);
		if(FAILED(hr)) return hr;
#ifdef NONEXCLUSIVE
		//hr=Real->BltFast(0, 0, menu, 0, 0);
		hr=Real->Blt(0, menu, 0, 0, 0);
#else
		//hr=Real->Blt(0, menu, 0, 0, 0);
		DDBLTFX ddbltfx;
		RECT r;
		ddbltfx.dwSize=sizeof(ddbltfx);
		ddbltfx.dwFillColor=0;
		r.left=   gHeight*4/3<gWidth ? (gWidth-gHeight *4/3)/2 : 0;
		r.top=    gHeight*4/3>gWidth ? (gHeight-gWidth *3/4)/2 : 0;
		r.right=  gWidth -r.left;
		r.bottom= gHeight-r.top;
		Real2->Blt(0, 0, 0, DDBLT_COLORFILL, &ddbltfx);
		Real2->Blt(&r, menu, 0, 0, 0);
		Real->BltFast(0,0,Real2,0,DDBLTFAST_WAIT);
#endif
		return DD_OK;
	} else {
		overlay->Unlock(a);
		return DD_OK;
	}
}
HRESULT _stdcall FakePrimarySurface::UpdateOverlay(LPRECT, LPDIRECTDRAWSURFACE4,LPRECT,DWORD, LPDDOVERLAYFX) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::UpdateOverlayDisplay(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::UpdateOverlayZOrder(DWORD, LPDIRECTDRAWSURFACE4) { UNUSEDFUNCTION; }
//*** Added in the v2 interface ***
HRESULT _stdcall FakePrimarySurface::GetDDInterface(LPVOID*) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::PageLock(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::PageUnlock(DWORD) { UNUSEDFUNCTION; }
//*** Added in the v3 interface ***
HRESULT _stdcall FakePrimarySurface::SetSurfaceDesc(LPDDSURFACEDESC2, DWORD) { UNUSEDFUNCTION; }
//*** Added in the v4 interface ***
HRESULT _stdcall FakePrimarySurface::SetPrivateData(REFGUID, LPVOID, DWORD, DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetPrivateData(REFGUID, LPVOID, LPDWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::FreePrivateData(REFGUID) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::GetUniquenessValue(LPDWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakePrimarySurface::ChangeUniquenessValue() { UNUSEDFUNCTION; }