#include "FakeCommon.h"
#include "SafeWrite.h"
#include "TexOverride.h"
#include "bitmapSaver.h"

#include <hash_map>
#include <vector>

using namespace std;
using namespace stdext;

static void SpawnTexture(DWORD width, DWORD height, DDSURFACEDESC2* desc, IDirectDrawSurface4** surf, IDirect3DTexture2** tex);
static void _cdecl CheckTex(void* tex);

#pragma pack(push)
#pragma pack(1)
struct tgaHeader
{
  BYTE IDLength;        /* 00h  Size of Image ID field */
  BYTE ColorMapType;    /* 01h  Color map type */
  BYTE ImageType;       /* 02h  Image type code */
  WORD CMapStart;       /* 03h  Color map origin */
  WORD CMapLength;      /* 05h  Color map length */
  BYTE CMapDepth;       /* 07h  Depth of color map entries */
  WORD XOffset;         /* 08h  X origin of image */
  WORD YOffset;         /* 0Ah  Y origin of image */
  WORD Width;           /* 0Ch  Width of image */
  WORD Height;          /* 0Eh  Height of image */
  BYTE PixelDepth;      /* 10h  Image pixel size */
  BYTE ImageDescriptor; /* 11h  Image descriptor byte */
};
#pragma pack(pop)

typedef void (_cdecl *RealFunc)(void*,void*);
typedef pair<DWORD, OverridenTex*> MyPair;
typedef pair<DWORD, IDirectDrawSurface4*> MyPair2;

static RealFunc rCopyTex=(RealFunc)0x0041D4F0;
static RealFunc rMipMap=(RealFunc)0x0041D680;

static bool Overriding=false;
static bool FirstOverride=false;
static OverridenTex* lastTex;

static hash_map<DWORD, OverridenTex*> texMap;
static hash_map<DWORD, IDirectDrawSurface4*> vidMap;
static vector<OverridenTex*> texList;

static OverridenTex* cTex=0;
OverridenTex* GetOverride() {
	return cTex;
}


static void _stdcall CreateVidTextureHook2(DWORD* d) {
	hash_map<DWORD, OverridenTex*> :: const_iterator itr;
	itr=texMap.find(*d);
	if(itr!=texMap.end()) {
		cTex=itr->second;
		return;
	} else {
		CheckTex((void*)d);
	}
	itr=texMap.find(*d);
	cTex=itr->second;
}

static void _declspec(naked) CreateVidTextureHook() {
	_asm {
		mov eax, esp;
		pushad;
		mov eax, [eax+4];
		mov eax, [eax];
		//mov eax, [eax];
		push eax;
		call CreateVidTextureHook2;
		popad;
		mov eax, 0x005BD320;
		jmp eax;
	}
}

static void ReadString(char* buf, const BYTE* ptr) {
	int index=0;
	while(*ptr!=1&&index<127) {
		buf[index++]=*ptr;
		ptr++;
	}
	buf[index]=0;
}

static void BltFromFile(char* dest, HANDLE source, DWORD width, DWORD height, DWORD destPitch, DWORD sourceDepth, bool invVert, bool invHorz) {
	DWORD unused;
	if(sourceDepth==24) {
		char* buf=new char[width*height*3 + 4];
		ReadFile(source, buf, width*height*3, &unused, 0);
		destPitch/=4;
		for(DWORD y=0;y<height;y++) {
			DWORD offset=(invVert?height-(y+1):y)*destPitch;
			DWORD offset2=y*width*3;
			if(invHorz) {
				for(DWORD x=0;x<width;x++) ((DWORD*)dest)[offset+(width-(x+1))]=(*(DWORD*)(&buf[offset2+x*3])) | 0xff000000;
			} else {
				for(DWORD x=0;x<width;x++) ((DWORD*)dest)[offset+x]=(*(DWORD*)(&buf[offset2+x*3])) | 0xff000000;
			}
		}
		delete[] buf;
	} else if(sourceDepth==32) {
		width <<= 2;
		if(invVert) {
			for(DWORD y=1;y<=height;y++) ReadFile(source, &dest[(height-y)*destPitch], width, &unused, 0);
		} else {
			if(destPitch==width) {
				ReadFile(source, dest, width*height, &unused, 0);
			} else {
				for(DWORD y=0;y<height;y++) ReadFile(source, &dest[y*destPitch], width, &unused, 0);
			}
		}
		if(invHorz) {
			DWORD tmp;
			for(DWORD y=0;y<height;y++) {
				for(DWORD x=0;x<width/2;x+=4) {
					tmp=*(DWORD*)&dest[y*destPitch+x];
					*(DWORD*)&dest[y*destPitch+x]=dest[y*destPitch+width-(x+4)];
					*(DWORD*)&dest[y*destPitch+width-(x+4)]=tmp;
				}
			}
		}
	}
}
static void GenMipMaps(IDirectDrawSurface4* from, const DDSURFACEDESC2* desc) {
	if(desc->dwMipMapCount) {
		DDSCAPS2 caps = { DDSCAPS_MIPMAP,0,0,0 };
		from->AddRef();
		IDirectDrawSurface4* to;
		for(DWORD i=1;i<desc->dwMipMapCount;i++) {
			from->GetAttachedSurface(&caps, &to);
			to->Blt(0, from, 0, DDBLT_WAIT, 0);
			from->Release();
			from=to;
		}
		to->Release();
	}
}
static bool LoadTga(const char* buf, OverridenTex* ot) {
	char buf2[MAX_PATH];
	strcpy_s(buf2, buf);
	strcat_s(buf2, ".tga");
	HANDLE file=CreateFileA(buf2, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
	if(file==INVALID_HANDLE_VALUE) return false;
	tgaHeader h;
	DWORD unused;
	ReadFile(file, &h, sizeof(h), &unused, 0);
	if((h.PixelDepth!=24&&h.PixelDepth!=32)||h.CMapLength!=0||h.ImageType!=2||h.ColorMapType!=0) {
		CloseHandle(file);
		return false;
	}

	ReadFile(file, buf2, h.IDLength, &unused, 0);
	DDSURFACEDESC2 desc;
	SpawnTexture(h.Width, h.Height, &desc, &ot->surf, &ot->tex);
	DDSURFACEDESC2 desc2;
	memset(&desc2, 0, sizeof(desc2));
	desc2.dwSize=sizeof(desc2);
	ot->surf->Lock(0, &desc2, DDLOCK_WRITEONLY|DDLOCK_WAIT, 0);
	BltFromFile((char*)desc2.lpSurface, file, h.Width, h.Height, desc2.lPitch, h.PixelDepth, (h.ImageDescriptor&0x20)==0, (h.ImageDescriptor&0x10)==0x10);
	ot->surf->Unlock(0);
	CloseHandle(file);
	GenMipMaps(ot->surf, &desc);
	ot->desc=desc;
	return true;
}
static bool LoadBmp(const char* buf, OverridenTex* ot) {
	char buf2[MAX_PATH];
	strcpy_s(buf2, buf);
	strcat_s(buf2, ".bmp");
	HANDLE file=CreateFileA(buf2, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
	if(file==INVALID_HANDLE_VALUE) return false;
	DWORD width, height, depth;
	if(!bmpLoad(file, &width, &height, &depth)) {
		CloseHandle(file);
		return false;
	}
	DDSURFACEDESC2 desc;
	SpawnTexture(width, height, &desc, &ot->surf, &ot->tex);
	DDSURFACEDESC2 desc2;
	memset(&desc2, 0, sizeof(desc2));
	desc2.dwSize=sizeof(desc2);
	ot->surf->Lock(0, &desc2, DDLOCK_WRITEONLY|DDLOCK_WAIT, 0);
	BltFromFile((char*)desc2.lpSurface, file, width, height, desc2.lPitch, depth, true, false);
	ot->surf->Unlock(0);
	CloseHandle(file);
	GenMipMaps(ot->surf, &desc);
	ot->desc=desc;
	return true;
}

static bool LoadDds(const char* buf, OverridenTex* ot) {
	char buf2[MAX_PATH];
	strcpy_s(buf2, buf);
	strcat_s(buf2, ".dds");
	HANDLE file=CreateFileA(buf2, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
	if(file==INVALID_HANDLE_VALUE) return false;
	DWORD magic, unused;
	ReadFile(file, &magic, 4, &unused, 0);
	if(magic!=0x20534444) {
		CloseHandle(file);
		return false;
	}
	DDSURFACEDESC2 desc;
	ReadFile(file, &desc, sizeof(desc), &unused, 0);
	if(desc.dwSize!=124||!(desc.ddpfPixelFormat.dwFlags&DDPF_RGB)||(desc.ddpfPixelFormat.dwRGBBitCount!=24&&desc.ddpfPixelFormat.dwRGBBitCount!=32)) {
		CloseHandle(file);
		return false;
	}
	if(desc.ddsCaps.dwCaps&DDSCAPS_ALPHA) desc.ddsCaps.dwCaps^=DDSCAPS_ALPHA;
	desc.ddsCaps.dwCaps|=DDSCAPS_SYSTEMMEMORY;
	DWORD depth=desc.ddpfPixelFormat.dwRGBBitCount;
	if(depth==24) {
		desc.ddpfPixelFormat.dwRGBBitCount=32;
		desc.ddpfPixelFormat.dwRBitMask=0xff0000;
		desc.ddpfPixelFormat.dwGBitMask=0xff00;
		desc.ddpfPixelFormat.dwBBitMask=0xff;
	}
	HRESULT hr=rDDraw->CreateSurface(&desc, &ot->surf, 0);
	DDSURFACEDESC2 desc2;
	memset(&desc2, 0, sizeof(desc2));
	desc2.dwSize=sizeof(desc2);
	IDirectDrawSurface4* surf=ot->surf;
	surf->AddRef();
	
	DWORD mipmaps;
	if(!(desc.dwFlags&DDSD_MIPMAPCOUNT)) mipmaps=1;
	else mipmaps=desc.dwMipMapCount;
	DDSCAPS2 caps = { DDSCAPS_MIPMAP,0,0,0 };
	for(DWORD i=1;i<=mipmaps;i++) {
		surf->Lock(0, &desc2, DDLOCK_WRITEONLY|DDLOCK_WAIT, 0);
		BltFromFile((char*)desc2.lpSurface, file, desc2.dwWidth, desc2.dwHeight, desc2.lPitch, depth, false, false);
		surf->Unlock(0);
		surf->Release();
		if(i<mipmaps) surf->GetAttachedSurface(&caps, &surf);
	}
	CloseHandle(file);

	ot->surf->QueryInterface(IID_IDirect3DTexture2, (void**)&ot->tex);
	ot->desc=desc;
	ot->desc.ddsCaps.dwCaps^=DDSCAPS_SYSTEMMEMORY;
	ot->desc.ddsCaps.dwCaps|=DDSCAPS_VIDEOMEMORY|DDSCAPS_ALLOCONLOAD;
	return true;
}

static void _cdecl CheckTex(void* tex) {
	BYTE* ptr=*(BYTE**)tex;
	if((DWORD)ptr<0x10000) goto end;
	BYTE* ptr2=ptr;
	for(int i=0;i<4;i++) if(ptr[i]!=(i+1)) goto end;
	for(int i=0;i<4;i++) if(ptr[i+4]!=(i+1)) goto end;
	for(int i=0;i<4;i++) if(ptr[i+9]!=(i+1)) goto end;
	if(ptr[8]!=1) goto end;
	char buf[MAX_PATH];
	ReadString(buf, &ptr[13]);
	if(strcmp(buf, "ddfix")) goto end;
	//This is definately an overridden texture
	strcpy_s(buf, TexturePath);
	ReadString(&buf[strlen(TexturePath)], &ptr[19]);
	OverridenTex* ot=new OverridenTex();
	if(!LoadTga(buf, ot)) {
		if(!LoadBmp(buf, ot)) {
			if(!LoadDds(buf, ot)) {
				delete ot;
				goto end;
			}
		}
	}
	lastTex=ot;
	texList.push_back(ot);
	texMap.insert(MyPair(*(DWORD*)tex, ot));
	return;
end:
	texMap.insert(MyPair(*(DWORD*)tex, 0));
}
static void _cdecl CopyTex(void* dest, void* source) {
	Overriding=false;
	BYTE* ptr=*(BYTE**)source;
	if(!ptr) goto end;
	BYTE* ptr2=ptr;
	for(int i=0;i<4;i++) if(ptr[i]!=(i+1)) goto end;
	for(int i=0;i<4;i++) if(ptr[i+4]!=(i+1)) goto end;
	for(int i=0;i<4;i++) if(ptr[i+9]!=(i+1)) goto end;
	if(ptr[8]!=1) goto end;
	char buf[128];
	ReadString(buf, &ptr[13]);
	if(strcmp(buf, "ddfix")) goto end;
	//This is definately an overridden texture
	strcpy_s(buf, TexturePath);
	ReadString(&buf[strlen(TexturePath)], &ptr[19]);
	OverridenTex* ot=new OverridenTex();
	if(!LoadTga(buf, ot)) {
		if(!LoadBmp(buf, ot)) {
			if(!LoadDds(buf, ot)) {
				delete ot;
				goto end;
			}
		}
	}
	lastTex=ot;
	Overriding=true;
	texList.push_back(ot);
	texMap.insert(MyPair((DWORD)dest, ot));
	rCopyTex(dest,source);
	return;
end:
	texMap.insert(MyPair((DWORD)dest, 0));
	rCopyTex(dest,source);
}

static void _stdcall LevelStartHook2() {
	for(DWORD i=0;i<texList.size();i++) {
		texList[i]->tex->Release();
		texList[i]->surf->Release();
	}
	texList.clear();
	texMap.clear();
	vidMap.clear();
}
static void _declspec(naked) LevelStartHook() {
	_asm {
		pushad;
		call LevelStartHook2;
		popad;
		mov eax, 0x0041C070;
		jmp eax;
	}
}
static void _cdecl MipMap(void* dest, void* source) {
	if(Overriding) {
		texMap.insert(MyPair(*(DWORD*)dest, lastTex));
	} else texMap.insert(MyPair(*(DWORD*)dest, 0));
	rMipMap(dest,source);
}
static void _declspec(naked) TextureExistsHook() {
	_asm {
		pushad;
		push eax;
		mov eax, [eax];
		call [eax+8];
		mov eax, [ebp+4];
		push eax;
		mov eax, [eax];
		call [eax+8];
		popad;
		mov eax, 0x005BD42B;
		jmp eax;
	}
}


/*static char ReadHookBuf[MAX_PATH];
static bool TwoCount=false;

static BOOL _stdcall ReadFileHook(HANDLE a, LPVOID b, DWORD c, LPDWORD d, LPOVERLAPPED e) {
	if(ReadHookBuf[0]) {
		if(!TwoCount) {
			TwoCount=true;
		} else {
			TwoCount=0;
			ReadHookBuf[0]=0;
		}
	}
	return ReadFile(a,b,c,d,e);
}*/

static HANDLE _stdcall CreateFileHook(const char* fname, DWORD access, DWORD share, LPSECURITY_ATTRIBUTES sec, DWORD creation, DWORD flags, HANDLE temp) {
	if(!_strcmpi(fname+strlen(fname)-4, ".pcx")||!_strcmpi(fname+strlen(fname)-4, ".gif")||!_strcmpi(fname+strlen(fname)-4, ".bmp")) {
		char buf[1024];
		strcpy_s(buf, fname);
		strcat_s(buf,".override");
		HANDLE h=CreateFileA(buf, access, share, sec, OPEN_EXISTING, flags, temp);
		if(h!=INVALID_HANDLE_VALUE) {
			//replaced=true;
			return h;
		}
		/*if(CheckForAuto) {
			const char* c=strstr(fname, "\\res\\");
			if(c&&strlen(fname<1024-15)) {
				memcpy(buf, fname, (DWORD)c-(DWORD)fname + 5);
				buf[(DWORD)c-(DWORD)fname + 6]=0;
				strcat_s(buf, "ddfix\\auto\\");
				strcat_s(buf, &c[5]);
				buf[strlen(buf)-4]=0;
				strcat_s(buf, ".tga");
				HANDLE h=CreateFileA(buf, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
				if(h==INVALID_HANDLE_VALUE) {
					buf[strlen(buf)-4]=0;
					strcat_s(buf, ".bmp");
					h=CreateFileA(buf, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
				}
				if(h==INVALID_HANDLE_VALUE) {
					buf[strlen(buf)-4]=0;
					strcat_s(buf, ".dds");
					h=CreateFileA(buf, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0);
				}
				if(h!=INVALID_HANDLE_VALUE) {
					TwoChar=false;
					strcpy_s(ReadHookBuf, "auto\\");
					strcat_s(ReadHookBuf, c+5);
					ReadHookBuf[strlen(ReadHookBuf)-4]=0;
				}
			}
		}*/
	}
	return CreateFileA(fname, access, share, sec, creation, flags, temp);
}

void OverrideSetup() {
	SafeWrite32(0x0041D948, (DWORD)&CopyTex - 0x0041D94C);
	SafeWrite32(0x0041D96B, (DWORD)&MipMap - 0x0041D96F);
	SafeWrite32(0x0062A67C, (DWORD)&CreateVidTextureHook);
	SafeWrite32(0x0040FCA4, (DWORD)&LevelStartHook - 0x0040FCA8);
	SafeWrite32(0x005BD427, (DWORD)&TextureExistsHook -  0x005BD42B);
	SafeWrite32(0x0060015C, (DWORD)&CreateFileHook);
	SafeWrite32(0x00591440, 0x000008c2);

	//SafeWrite16(0x0041D95F, 0x9090);
	//ReadHookBuf[0]=0;
	//SafeWrite32(0x006001A4, (DWORD)&ReadFileHook);
}

static void SpawnTexture(DWORD width, DWORD height, DDSURFACEDESC2* desc, IDirectDrawSurface4** surf, IDirect3DTexture2** tex) {
	*surf=0;
	*tex=0;
	memset(desc, 0, sizeof(DDSURFACEDESC2));
	desc->dwSize=sizeof(DDSURFACEDESC2);
	desc->dwFlags=DDSD_CAPS|DDSD_WIDTH|DDSD_HEIGHT|DDSD_PIXELFORMAT;
	desc->dwWidth=width;
	desc->dwHeight=height;
	desc->ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	desc->ddpfPixelFormat.dwFlags=DDPF_ALPHAPIXELS|DDPF_RGB;
	desc->ddpfPixelFormat.dwRGBBitCount=32;
	desc->ddpfPixelFormat.dwBBitMask=0xff;
	desc->ddpfPixelFormat.dwGBitMask=0xff00;
	desc->ddpfPixelFormat.dwRBitMask=0xff0000;
	desc->ddpfPixelFormat.dwRGBAlphaBitMask=0xff000000;
	desc->ddsCaps.dwCaps=DDSCAPS_TEXTURE|DDSCAPS_SYSTEMMEMORY;

	DWORD mipmaps=0;
	DWORD test=(width>height)?width:height;
	while(test>1) {
		mipmaps++;
		test >>= 1;
	}
	if(mipmaps) {
		desc->dwFlags|=DDSD_MIPMAPCOUNT;
		desc->ddsCaps.dwCaps|=DDSCAPS_COMPLEX|DDSCAPS_MIPMAP;
		desc->dwMipMapCount=mipmaps;
	}

	rDDraw->CreateSurface(desc, surf, 0);
	(*surf)->QueryInterface(IID_IDirect3DTexture2, (void**)tex);

	desc->ddsCaps.dwCaps^=DDSCAPS_SYSTEMMEMORY|DDSCAPS_VIDEOMEMORY|DDSCAPS_ALLOCONLOAD;
}

IDirectDrawSurface4* IsTextureInVRam(OverridenTex* ot) {
	hash_map<DWORD, IDirectDrawSurface4*> :: const_iterator itr;
	itr=vidMap.find((DWORD)ot);
	if(itr==vidMap.end()) return 0;
	else return itr->second;
}
void AddTextureToVRam(OverridenTex* ot, IDirectDrawSurface4* surf) {
	vidMap.insert(MyPair2((DWORD)ot, surf));
}
void RemoveTextureFromVRam(OverridenTex* ot) {
	vidMap.erase((DWORD)ot);
}